#include "pkgs_source.h"
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <rpm/rpmio.h>
#include <curl/curl.h>
	
#include "utils.h"
#include "../cfile.h"

Tps_set PsSet;

int getFileFromURL(int protocol,char *url, char *dest, char *destDir);

int psOpenInternal(int protocol, char *server, char *name,char *path, int rm)
{
	char src[1000],dest[1000],destdir[1000];
	int ret;
	int rc;
	switch(protocol){
		case PROTOCOL_FTP:
		case PROTOCOL_EXTERNAL:
			sprintf(src,"ftp://%s/%s",server,name);
			sprintf(dest,"/%s/ftp/%s/%s",getVar(),server,name);
			strcpy(destdir,dest);
			*strrchr(destdir,'/')='\0';
			if(path)
				strcpy(path,dest);
			log_debug("trying from cache file %s",dest);
		
			ret=open(dest,O_RDONLY);
			if(ret>0){
				log_debug("opened from cache");
				return ret;
			}
			remMultiSl(destdir);
			remMultiSl(dest);
			log_debug("making dir (%s) to cache file",destdir);
			mkdirhier(destdir);
			log_debug("fetching from ftp file #%s#",src);
			rc=getFileFromURL(protocol,src,dest,destdir);
			if(rc<0)
			        log_message(
			          _("FTP error: can't get file %s "),src);
			rc =open(dest,O_RDONLY);
			if (rc >= 0 && rm) 
				unlink(dest);
				/* file will remain on this
				 * until filehandle is closed */
			return rc;
		case PROTOCOL_DIR:
			return open(name,O_RDONLY);
	}
	return -1;
}

int psCacheRemoveInfo(Tpkgs_source *ps,char *name)
{
	char buf[1000];
	log_debug("Removing cache file %s",name);
	if(ps->protocol!=PROTOCOL_FTP)
		return 0;
	snprintf(buf,999,"%s/ftp/%s/%s/%s",getVar(),ps->server,
			ps->infoDir,name);
	log_debug("unlinking file %s",buf);
	unlink(buf);

	snprintf(buf,999,"%s/ftp/%s/%s/%s.gz",getVar(),ps->server,
			ps->infoDir,name);
	log_debug("unlinking file %s",buf);
	unlink(buf);
	
	snprintf(buf,999,"/%s/wuch-cache/ftp/%s/%s/%s.gz",getTmp(),ps->server,
			ps->infoDir,name);
	log_debug("unlinking file %s",buf);
	unlink(buf);
	log_debug("deletion finisched");
	
	
	return 1;
}
		
extern int remove_cache;

#if 0
struct ps_cache {
	struct ps_cache *next;
	int fd;
	char *path;
};

static struct ps_cache *ps_head;


void psClosePkg(int fd)
{
	struct ps_cache *self = 0, *p;
	
	if (ps_head && ps_head->fd == fd) {
		self = ps_head;
		ps_head = ps_head->next;
	} else if (ps_head) {
		for (p = ps_head; p->next; p = p->next) 
			if (p->next->fd == fd)
				break;
		if (p->next) {
			self = p->next;
			p->next = self->next;
		}
	}

	close(fd);

	if (self) {
		if (remove_cache) 
			unlink(self->path);
		free(self->path);
		free(self);
	}
}
#else
void psClosePkg(int fd)
{
	close(fd);
}
#endif

void psFetchPkg(Tpkgs_source *ps, char *name)
{
	if(ps->protocol == PROTOCOL_FTP)
		close(psOpenPkg(ps,name));	
}

int psOpenPkg(Tpkgs_source *ps, char *name)
{
	char full[1000];
	int fd;
	struct ps_cache *p;
	
	sprintf(full, "/%s/%s",ps->pkgsDir,name);
	switch (ps->protocol){
		case PROTOCOL_FTP:
		case PROTOCOL_EXTERNAL:
			fd = psOpenInternal(ps->protocol,ps->server,
					    full,NULL,remove_cache);
		#if 0
			if (fd >= 0) {
				p = malloc(sizeof(*p));
				p->fd = fd;
				p->path = strdup(full);
				p->next = ps_head;
				ps_head = p;
			}
		#endif
			return fd;
		case PROTOCOL_DIR:
			return open(full,O_RDONLY);
		default:
			fprintf(stderr,
				_("THIS PROTOCOL ISn't IMPLEMENTED yet\n"));
			break;
	}
	return -1;  /* error */
}
			
int psOpenInfo(Tpkgs_source *ps, char *name){
	char full[1000];
	char local[1000],sys[1000],tmpf[1000];
	int ret;

	log_debug("opening info file %s",name);
	
	if(strcmp("tocfile.md5",name)==0)
		psCacheRemoveInfo(ps,name);

	sprintf(full, "/%s/%s",ps->infoDir,name);
	switch (ps->protocol){
		case PROTOCOL_FTP:
		case PROTOCOL_EXTERNAL:
			log_debug("calling OpenInternal");
			ret=psOpenInternal(ps->protocol,ps->server,
					    full,local,0);
			if(ret!=-1){
				log_debug("file %s opened",name);
				return ret;
			}
			strcat(full,".gz");
			ret=psOpenInternal(ps->protocol,ps->server,
					   full,local,0);
			if(ret==-1){
				log_message("There is no way to open info file %s  !",full);
				return ret;
			}
			sprintf(sys,"gunzip %s ",local);
			if(system(sys)){
				log_message("system(%s) failed",sys);
				return -1;
			}
			*strrchr(local,'.')='\0';
			log_message("file %s fetched and ungziped",full);
			return open(local,O_RDONLY);
			
			
		case PROTOCOL_DIR:
			ret=open(full,O_RDONLY);
			if(ret!=-1)
				return ret;
			strcat(full,".gz");
			tmpnam(tmpf);
			sprintf(sys,"gunzip -c %s > %s",full,tmpf);
			system(sys);
			ret = open(tmpf, O_RDONLY);
			log_message("Unlinking %s status: %d", tmpf,
					unlink(tmpf));
			return ret;
		default:
			fprintf(stderr,
				_("THIS PROTOCOL ISn't IMPLEMENTED yet\n"));
			break;
	}
	return -10;  /* error */

};

void psSetInit()
{
	int i,j;
	char *buf,b1[1000];
	char *tmp,*prefix;

	PsSet.count=0;
	PsSet.ps=NULL;

	i=0;
	j=0;
	while((buf=getSource(j++))){
		PsSet.ps=realloc(PsSet.ps,sizeof(Tpkgs_source)*(i+1));
		PsSet.ps[i].server=NULL;
		
		tmp=strtok(buf,",");
		if(tmp == NULL){
			log_message("Bad formed source specyfication !");
			continue;
		}
		PsSet.ps[i].name=strdup(tmp);
		
		tmp=strtok(NULL,",");
		if(tmp == NULL){
			log_message("Bad formed source specyfication !");
			continue;
		}

			
		PsSet.ps[i].protocol=PROTOCOL_UNKNOWN;
		if(strcasecmp(tmp,"ftp")==0){
			PsSet.ps[i].protocol=PROTOCOL_FTP;
			tmp=strtok(NULL,",");
			PsSet.ps[i].server=strdup(tmp);
		}	
		
		if(strcasecmp(tmp,"external")==0){
			PsSet.ps[i].protocol=PROTOCOL_EXTERNAL;
			tmp=strtok(NULL,",");
			PsSet.ps[i].server=strdup(tmp);
		}

		if(strcasecmp(tmp,"dir")==0){
			PsSet.ps[i].protocol=PROTOCOL_DIR;
		}
		if(strcasecmp(tmp,"cdrom")==0){
			PsSet.ps[i].protocol=PROTOCOL_CDROM;
		}

		if(PsSet.ps[i].protocol==PROTOCOL_UNKNOWN){
			fprintf(stderr,_("Unknown protocol type '%s' !\n"),tmp);
			continue;
		}

		tmp=strtok(NULL,",");
		if(tmp == NULL){
			log_message("Bad formed source specyfication !");
			continue;
		}
		prefix=strdup(tmp);

		tmp=strtok(NULL,",");
		if(tmp == NULL){
			log_message("Bad formed source specyfication !");
			continue;
		}
		sprintf(b1,"/%s/%s",prefix,tmp);
		PsSet.ps[i].pkgsDir=strdup(b1);

		tmp=strtok(NULL,",");
		if(tmp == NULL){
			log_message("Bad formed source specyfication !");
			continue;
		}
		sprintf(b1,"/%s/%s",prefix,tmp);
		PsSet.ps[i].infoDir=strdup(b1);

		free(prefix);	
		i++;
	}
	PsSet.count=i;
}




int getFileFromURLCurl(char *url, char *dest)
{
	static CURL *ch=NULL;
	FILE *out;
	int rc;
	char errorMsg[CURL_ERROR_SIZE+1];

	if(ch == NULL){
		ch = curl_easy_init();
		if(!ch){
			log_message("Can't init libcurl !");
			return -1;
		}else{
			log_message("Curl initialized");
		}
	}
	rc = unlink(dest);
	log_message("%s unlink status: %d (%s) ", dest, rc, strerror(errno));
	out = fopen(dest,"w");
	if(!out){
		log_message("%s: Can't open dest file",__FUNCTION__);
		return -2;
	}
	log_message("setting curl's options");

	if(curl_easy_setopt(ch, CURLOPT_FILE, out)){
		log_message("curl_easy_setopt can't set FILE");
		fclose(out);
		return -3;
	}

	if(curl_easy_setopt(ch, CURLOPT_URL,  url)){
		log_message("curl_easy_setopt can't set URL");
		fclose(out);
		return -4;
	};
	if(curl_easy_setopt(ch, CURLOPT_NOPROGRESS,1)){
		log_message("curl_easy_setopt can't disable progress status");
		fclose(out);
		return -5;
	}
	if(curl_easy_setopt(ch, CURLOPT_ERRORBUFFER,errorMsg)){
		log_message("curl_easy_setopt can't set error msg buffer");
		fclose(out);
		return -5;
	}
	if(curl_easy_setopt(ch, CURLOPT_RESUME_FROM,0)){
		log_message("curl_easy_setopt can't set \"resume from\"");
		fclose(out);
		return -5;
	}




	log_message("fetching file...");
	if(curl_easy_perform(ch)){
		log_message("curl can't fetch file (%s)",url);
		fclose(out);
		log_message("%s unlink status: %d ",dest,unlink(dest));
		return -6;
	}else{
		log_message("file (%s) fetched",url);
	}
	fclose(out);
	/* curl_easy_cleanup(ch); */
	return 0;
}

int getFileFromURL(int protocol,char *url, char *dest, char *destDir)
{
	if(protocol==PROTOCOL_FTP)
		return getFileFromURLCurl(url,dest);
	if(protocol==PROTOCOL_EXTERNAL){
		char *command;
		int rc;
		command = getFetcherCommand(url, destDir);
		if(command == NULL){
			log_message("FATAL: fetcher not set! check your config file\n");
			exit(1);
		}
		log_message("executing '%s'",command);
		rc=system(command);
		if(rc){
			log_message("command failed");
			return -10;
		}
		return 0;
	}
	log_message("internal error #45. report to <pawelk@pld.org.pl>");
	exit(1);
}
